
#include <8051io.h>
#include <8051reg.h>


#define FALSE 0
#define TRUE 255

#define word unsigned int
#define byte unsigned char

#define SAME 0
#define OPPOSITE 1
#define DIT 2
#define DAH 3


register byte pad,lastPad;      // Pad state, 0=idle 1=dit 2=dah System record last pad state
register word trig1,trig2;      // Current pads trigger value
register word tFrq;             // Sidetone frequency (system tick)
register byte tone_on;          // Sidetone enabling flag
register word count;            // Timer count
register byte CWspd;            // Code speed (timer tick)
register byte CWFspd;           // Code speed in function entry (timer tick)
register word PadResync;        // After 30sec of pad idle, pad1/pad2 are reloaded

// Configuration
//---------------
register word SideTone;         // Tone frequency
register byte Wpm;              // Operating code speed
register byte WpmFunc;          // Function entry code speed (not used for now)
register byte WpmPratice;       // Pratice CW letter speed
register byte InvertPad;        // Invert paddle flag
register byte PadPriority;      // Paddle priority: (0)same, (1)opposite, (2)dit or (3)dah
register byte ToneEnable;       // Sidetone enable flag

register byte FuncMode;         // Function mode, flag


// This table start at code 33 of standard ascii
// The format of table is simple, 1st caracter is number of element (dit or dah) of
// caracter and second is caracter, dot=0 dah=1



char CWcode[] = { \
        0,0x00, \      // !
        6,0x48, \      // "
        0,0x00, \      // #
        7,0x12, \      // $
        0,0x00, \      // %
        0,0x00, \      // &
        6,0x78, \      // '
        5,0xB0, \      // (
        6,0xB4, \      // )
        0,0x00, \      // *
        5,0x50, \      // +
        6,0xCC, \      // ,
        6,0x84, \      // -
        6,0x54, \      // .
        5,0x90, \      // /
        5,0xF8, \      // 0
        5,0x78, \      // 1
        5,0x38, \      // 2
        5,0x18, \      // 3
        5,0x08, \      // 4
        5,0x00, \      // 5
        5,0x80, \      // 6
        5,0xC0, \      // 7
        5,0xE0, \      // 8
        5,0xF0, \      // 9
        6,0xE0, \      // :
        6,0xA8, \      // ;
        0,0x00, \      // <
        5,0x88, \      // =
        0,0x00, \      // >
        6,0x30, \      // ?
        0,0x00, \      // @
        2,0x40, \      // A
        4,0x80, \      // B
        4,0xA0, \      // C
        3,0x80, \      // D
        1,0x00, \      // E
        4,0x20, \      // F
        3,0xC0, \      // G
        4,0x00, \      // H
        2,0x00, \      // I
        4,0x70, \      // J
        3,0xA0, \      // K
        4,0x40, \      // L
        2,0xC0, \      // M
        2,0x80, \      // N
        3,0xE0, \      // O
        4,0x60, \      // P
        4,0xD0, \      // Q
        3,0x40, \      // R
        3,0x00, \      // S
        1,0x80, \      // T
        3,0x20, \      // U
        4,0x10, \      // V
        3,0x60, \      // W
        4,0x90, \      // X
        4,0xB0, \      // Y
        4,0xC0, \      // Z
        0,0x00, \      // [
        0,0x00, \      // \
        0,0x00, \      // ]
        0,0x00, \      // ^
        6,0x34, \      // _
        0,0x00 };      // `



void timer1int() asm {
t1a:
        org     ?ROM+1Bh
        ajmp    t1a

        org     t1a
        mov     TL1,tFrq                // Reload timer with frequency
        mov     TH1,tFrq+1
        push    ACC

        mov     A,tone_on               // Check if speaker is enabled
        jz      t1b
        cpl     P1.1                    // Toggle speaker pin

t1b:    inc     count                   // Increment counter
        mov     A,count
        jnz     t1c
        inc     count+1

t1c:    pop     ACC
        reti
}



void GetPadState() {
    word val1,val2;

    asm {
TR0     BIT     8Ch                 Timer run flag
TR1     BIT     8Eh

        clr     P1.0                Enable PAD #1
        mov     TL0,#0              Clear counter
        mov     TH0,#0
        setb    TR0                 Enable gate
}
        delay(5);                   // Wait 5ms
asm {
        clr     TR0                 Disable gate
}
        val1=TL0+(TH0<<8);          // Get counter result

asm {
        setb    P1.0                Enable PAD #2
        mov     TL0,#0              Clear counter
        mov     TH0,#0
        setb    TR0                 Enable gate
}
        delay(5);                   // Wait 5ms
asm {
        clr     TR0                 Disable gate
}
        val2=TL0+(TH0<<8);          // Get counter result

    pad=0;
    if(val1<trig1) pad=1;           // Get raw paddles state
    if(val2<trig2) pad+=2;

    if(pad==0) {                    // After 30sec of inactivity, reload trigging level
        if(PadResync>3000) {
            trig1 = val1*19/20;
            trig2 = val2*19/20;
            PadResync=0;
        }
        PadResync++;
        lastPad=0;
        return;
    } else PadResync=0;

    if(InvertPad) {                 // Invert paddles if need
        pad=(~pad)&3;
        if(pad==0) pad=3;
    }

    if(pad==3) {                    // Use priority to determine pad, when both are pressed
        if(lastPad==0) lastPad=1;
        switch(PadPriority) {
            case SAME:      pad=lastPad; break;
            case OPPOSITE:  pad=(~lastPad)&3; break;
            case DIT:       pad=1; break;
            case DAH:       pad=2; break;
        }
        return;
    }

    lastPad=pad;
}




void SetWPM() {
    word tmp;

    tmp = ((word)Wpm*100) / 12;              // Convert wpm to Hz
    CWspd = (SideTone*10) / tmp;             // Find value in count tick

    tmp = ((word)WpmFunc*100) / 12;
    CWFspd = (SideTone*10) / tmp;
}


void SetTone() {
    tFrq = ~(50000 / (SideTone/20));

    SetWPM();                     // resync dit/dah length value
}




//
// Wait a CW unit time, check if it for sending or for function entry
//
void WaitUnit(char unit) {
    byte i;

    for(i=0; i<unit; i++) {
        count = 0;
        if(FuncMode) while(count<CWFspd); else while(count<CWspd);
    }
}


//
// Function mode: ptt line is disable and sidetone are enabled.
// Normal mode: line are keyed and sidetone are used if need.
//
void Dit() {
    if(FuncMode) {
        tone_on=TRUE;
        WaitUnit(1);
        tone_on=FALSE;
        WaitUnit(1);
    } else {
        if(ToneEnable) tone_on=TRUE;
        asm {
        setb P1.2
        }
        WaitUnit(1);
        asm {
        clr P1.2
        }
        tone_on=FALSE;
        WaitUnit(1);
    }
}



//
// if ptt=0, ptt line is disable and sidetone are enabled (function mode)
// if ptt=1, ptt line are keyed and sidetone are used if need (normal operation)
//
void Dah() {
    if(FuncMode) {
        tone_on=TRUE;
        WaitUnit(3);
        tone_on=FALSE;
        WaitUnit(1);
    } else {
        if(ToneEnable) tone_on=TRUE;
        asm {
        setb P1.2
        }
        WaitUnit(3);
        asm {
        clr P1.2
        }
        tone_on=FALSE;
        WaitUnit(1);
    }
}


void SendCW(byte car) {
    byte length,code;

    if(car>32 && car<97) {
        car-=33;
        length = CWcode[car*2];
        code   = CWcode[(car*2)+1];
        if(length==0) return;
        for(car=0; car<length; car++) {
            if(code&0x80) Dah(); else Dit();
            code<<=1;
        }
        WaitUnit(3);
    }
}

//
// Return 0 if no caracter sended, 1 if bad caracter. return ASCII caracter.
//
byte GetCW() {
    word limit;
    byte length,code;

    length=code=0;              // Reset variable
    limit=(word)CWFspd*2;
    count=0;

    while(count<limit) {
        GetPadState();
        if(pad) {
            code<<=1;
            if(pad&2) { code|=1; Dah(); } else Dit();
            length++;
            count=0;
        }
    }
    if(length==0) return 0;

    code<<=(8-length);
    for(limit=0; limit<128; limit+=2)
        if(CWcode[limit]==length && CWcode[limit+1]==code) return (limit/2)+33;

    return 1;
}





void F_Error() {
     byte i;

    for(i=0; i<50; i++) {
        tone_on=TRUE;
        count=0; while(count<5);
        tone_on=FALSE;
        count=0; while(count<5);
    }
}


void F_OK() {
    word tmp;

    tmp=SideTone;
    SideTone = 1400;
    SetTone();
    Dah();
    SideTone = tmp;
    SetTone();
}



void F_ToneFreq() {
    byte d1,d2;

    do d1=GetCW(); while(d1==0);
    do d2=GetCW(); while(d2==0);
    d1-=48; d2-=48;
    if(d1>=5 && d1<=9)
        if(d1>=0 && d1<=9) {
            SideTone = ((word)d1*100) + ((word)d2*10);
            SetTone();
            F_OK();
            return;
        }
    F_Error();
}

void F_WPM() {
    byte d1,d2;

    do d1=GetCW(); while(d1==0);
    do d2=GetCW(); while(d2==0);

    if(d2<'0' || d2>'9') { F_Error(); return; }
    d2-=48;

    switch(d1) {
        case 'U': if((Wpm+d2)>60) { F_Error(); return; }
                  Wpm+=d2;
                  SetWPM();
                  break;
        case 'D': d1 = Wpm-d2;
                  if(d1<6 || d1>60) { F_Error(); return; }
                  Wpm=d1;
                  SetWPM();
                  break;
        default:  if(d1<'0' || d1>'9') { F_Error(); return; }
                  d1-=48;
                  d1 = (d1*10)+d2;
                  if(d1<6 || d1>60) { F_Error(); return; }
                  Wpm=d1;
                  SetWPM();
                  break;
    }
    F_OK();
}


void F_WPMF() {
    byte d1,d2;

    do d1=GetCW(); while(d1==0);
    do d2=GetCW(); while(d2==0);

    if(d2<'0' || d2>'9' || d1<'0' || d1>'9') { F_Error(); return; }
    d1-=48; d2-=48;
    d1 = (d1*10)+d2;
    if(d1<6 || d1>30) { F_Error(); return; }
    WpmFunc=d1;
    SetWPM();
    F_OK();
}


void F_InvPad() {
    InvertPad=~InvertPad;
    F_OK();
}


void F_Priority() {
    byte car;

    do car=GetCW(); while(car==0);

    switch(car) {
        case 'S':   PadPriority=0; break;
        case 'O':   PadPriority=1; break;
        case 'I':   PadPriority=2; break;
        case 'A':   PadPriority=3; break;
        default:    F_Error(); return;
    }

    F_OK();
}


void F_Monitor() {
    ToneEnable=~ToneEnable;
    F_OK();
}


void F_Tune() {
    asm {
    setb P1.2
    }
    do GetPadState(); while(pad==0);
    asm {
    clr P1.2
    }
    F_OK();
}

void F_Hand() {
    while((P1&0xF0)==0xF0) {
        GetPadState();
        if(pad) {
            if(ToneEnable) tone_on=TRUE;
            asm {
            setb P1.2
            }
        } else {
            asm {
            clr P1.2
            }
            tone_on=FALSE;

        }
    }
    F_OK();
}

void F_SetPratice() {
    byte d1,d2;

    do d1=GetCW(); while(d1==0);
    do d2=GetCW(); while(d2==0);

    if(d2<'0' || d2>'9' || d1<'0' || d1>'9') { F_Error(); return; }
    d1-=48; d2-=48;
    d1 = (d1*10)+d2;
    if(d1<5 || d1>60) { F_Error(); return; }
    WpmPratice=d1;
    F_OK();
}


void F_Pratice() {
    byte car;
    word CWdly;
    word OldWpm;

    CWdly = ((word)WpmPratice*100) / 12;       // Convert wpm to Hz
    CWdly = (SideTone*10) / CWdly;             // Find value in count tick

    if(CWdly<CWspd) { F_Error(); return; }     // Operating speed cannot be inferior to pratice speed
    CWdly*=3;                                  // Caracter spacing is 3 unit
    CWdly-=CWspd*3;                            // Substract with 3 unit of operating speed

    OldWpm = WpmFunc;                          // Replace WPM of function by WPM of operation
    WpmFunc = Wpm;
    SetWPM();

    while((P1&0xF0)==0xF0) {
        car=rand(91);
        if(car<44) continue;
        if(car>57 && car<63) continue;
        SendCW(car);
        if(CWdly) { count=0; while(count<CWdly); }
    }

    WpmFunc=OldWpm;
    SetWPM();
    F_OK();
}


void FunctionMode() {
    byte car;

    FuncMode = TRUE;
    SendCW('F');
    do car=GetCW(); while(car==0);

    switch(car) {
        case 'T': F_ToneFreq(); break;
        case 'S': F_WPM(); break;
        case 'F': F_WPMF(); break;
        case 'R': F_InvPad(); break;
        case 'P': F_Priority(); break;
        case 'M': F_Monitor(); break;
        case 'X': F_Tune(); break;
        case 'H': F_Hand(); break;
        case 'C': F_SetPratice(); break;
        case 'W': F_Pratice(); break;
        default:  F_Error();
    }

    FuncMode = FALSE;
}



void main() {
    byte c;

    delay(500);                 // Wait stabilisation of power and 556 oscillator.

    TCON = 0x40;                // Run timer #1
    TMOD = 0x15;                // T0 16bit counter, T1 16bit timer
    IE = 0x88;                  // Enable MIE and T1 interrupt
    P1 = 0xF8;


    // SETUP DEFAULT CONFIG
    //----------------------
    SideTone = 750;             // Sidetone = 750hz
    tone_on = FALSE;
    SetTone();
    ToneEnable = TRUE;          // Sidetone enabled
    Wpm = 20;                   // CW speed = 20wpm
    WpmFunc = 20;               // CW function speed = 20wpm
    WpmPratice = 8;             // CW pratice speed = 8wpm
    SetWPM();
    InvertPad = FALSE;          // Paddle in normal order
    PadPriority = OPPOSITE;     // Priority to opposite paddle
    PadResync = -1;             // Reload pad value
    trig1=trig2=0;

    FuncMode = TRUE;
    SendCW('O');
    SendCW('K');
    FuncMode = FALSE;

    // MAIN LOOP
    //-----------
    while(1) {
        
        GetPadState();
        if(pad&2) { Dah(); PadResync=0; continue; }
        if(pad&1) { Dit(); PadResync=0; continue; }

        
        if( (P1&0xF0)!=0xF0) {
            delay(500);             // Wait, if 2 or more button was pressed in same time
            c = (~P1) >> 4;
            switch(c) {
                case 1+2: FunctionMode(); break;
            }
        }

    }
}
